@@ -5,16 +5,16 @@ import numpy
 from ._common import num_nodes_per_cell
 from ._exceptions import ReadError, WriteError
 from ._files import is_buffer
-from ._mesh import Mesh
+from ._mesh import CellBlock, Mesh
 
-_extension_to_filetype = {}
+extension_to_filetype = {}
 reader_map = {}
 _writer_map = {}
 
 
 def register(name, extensions, reader, writer_map):
     for ext in extensions:
-        _extension_to_filetype[ext] = name
+        extension_to_filetype[ext] = name
 
     if reader is not None:
         reader_map[name] = reader
@@ -25,9 +25,9 @@ def _filetype_from_path(path):
     ext = ""
     out = None
     for suffix in reversed(path.suffixes):
-        ext = suffix + ext
-        if ext in _extension_to_filetype:
-            out = _extension_to_filetype[ext]
+        ext = (suffix + ext).lower()
+        if ext in extension_to_filetype:
+            out = extension_to_filetype[ext]
 
     if out is None:
         raise ReadError("Could not deduce file format from extension '{}'.".format(ext))
@@ -47,7 +47,8 @@ def read(filename, file_format=None):
             raise ReadError("File format must be given if buffer is used")
         if file_format == "tetgen":
             raise ReadError(
-                "tetgen format is spread across multiple files, and so cannot be read from a buffer"
+                "tetgen format is spread across multiple files "
+                "and so cannot be read from a buffer"
             )
         msg = "Unknown file format '{}'".format(file_format)
     else:
@@ -74,13 +75,23 @@ def write_points_cells(
     point_data=None,
     cell_data=None,
     field_data=None,
+    point_sets=None,
+    cell_sets=None,
     file_format=None,
     **kwargs
 ):
     points = numpy.asarray(points)
-    cells = {key: numpy.asarray(value) for key, value in cells.items()}
+    if isinstance(cells, dict):
+        cells = [CellBlock(name, vals) for name, vals in cells.items()]
+    cells = [(key, numpy.asarray(value)) for key, value in cells]
     mesh = Mesh(
-        points, cells, point_data=point_data, cell_data=cell_data, field_data=field_data
+        points,
+        cells,
+        point_data=point_data,
+        cell_data=cell_data,
+        field_data=field_data,
+        point_sets=point_sets,
+        cell_sets=cell_sets,
     )
     return write(filename, mesh, file_format=file_format, **kwargs)
 
@@ -110,14 +121,14 @@ def write(filename, mesh, file_format=None, **kwargs):
     try:
         writer = _writer_map[file_format]
     except KeyError:
-        raise KeyError(
+        raise WriteError(
             "Unknown format '{}'. Pick one of {}".format(
                 file_format, sorted(list(_writer_map.keys()))
             )
         )
 
     # check cells for sanity
-    for key, value in mesh.cells.items():
+    for key, value in mesh.cells:
         if key[:7] == "polygon":
             if value.shape[1] != int(key[7:]):
                 raise WriteError()
