@@ -3,16 +3,17 @@ I/O for Gmsh's msh format (version 4.0, as used by Gmsh 4.1.5), cf.
 <http://gmsh.info//doc/texinfo/gmsh.html#MSH-file-format-_0028version-4_0029>.
 """
 import logging
-import struct
 from functools import partial
 
 import numpy
 
-from .._common import cell_data_from_raw, raw_from_cell_data
+from .._common import _geometric_dimension, cell_data_from_raw, raw_from_cell_data
 from .._exceptions import ReadError, WriteError
-from .._mesh import Mesh
+from .._mesh import CellBlock, Mesh
 from .common import (
+    _gmsh_to_meshio_order,
     _gmsh_to_meshio_type,
+    _meshio_to_gmsh_order,
     _meshio_to_gmsh_type,
     _read_data,
     _read_physical_names,
@@ -30,7 +31,6 @@ c_double = numpy.dtype("d")
 def read_buffer(f, is_ascii, data_size):
     # Initialize the optional data fields
     points = []
-    cells = {}
     field_data = {}
     cell_data_raw = {}
     cell_tags = {}
@@ -102,7 +102,7 @@ def _read_entities(f, is_ascii, data_size):
     for d, n in enumerate(number):
         for _ in range(n):
             tag = int(fromfile(f, c_int, 1)[0])
-            fromfile(f, c_double, 6)  # discard boxMinX…boxMaxZ
+            fromfile(f, c_double, 6)  # discard boxMinX...boxMaxZ
             num_physicals = int(fromfile(f, c_ulong, 1)[0])
             physical_tags[d][tag] = list(fromfile(f, c_int, num_physicals))
             if d > 0:  # discard tagBREP{Vert,Curve,Surfaces}
@@ -176,7 +176,6 @@ def _read_elements(f, point_tags, physical_tags, is_ascii, data_size):
     num_entity_blocks, total_num_elements = fromfile(f, c_ulong, 2)
 
     data = []
-    cell_data = {}
 
     for k in range(num_entity_blocks):
         # tagEntity(int) dimEntity(int) typeEle(int) numElements(unsigned long)
@@ -210,35 +209,15 @@ def _read_elements(f, point_tags, physical_tags, is_ascii, data_size):
     # Note that the first column in the data array is the element tag; discard it.
     data = [(physical_tag, tpe, itags[d[:, 1:]]) for physical_tag, tpe, d in data]
 
-    cells = {}
+    cells = []
+    cell_data = {"gmsh:physical": []}
     for physical_tag, key, values in data:
-        if key in cells:
-            cells[key] = numpy.concatenate([cells[key], values])
-            if physical_tag:
-                if key not in cell_data:
-                    cell_data[key] = {}
-                cell_data[key]["gmsh:physical"] = numpy.concatenate(
-                    [
-                        cell_data[key]["gmsh:physical"],
-                        physical_tag[0] * numpy.ones(len(values), int),
-                    ]
-                )
-        else:
-            cells[key] = values
-            if physical_tag:
-                if key not in cell_data:
-                    cell_data[key] = {}
-                cell_data[key]["gmsh:physical"] = physical_tag[0] * numpy.ones(
-                    len(values), int
-                )
-
-    # Gmsh cells are mostly ordered like VTK, with a few exceptions:
-    if "tetra10" in cells:
-        cells["tetra10"] = cells["tetra10"][:, [0, 1, 2, 3, 4, 5, 6, 7, 9, 8]]
-    if "hexahedron20" in cells:
-        cells["hexahedron20"] = cells["hexahedron20"][
-            :, [0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 16, 9, 17, 10, 18, 19, 12, 15, 13, 14]
-        ]
+        cells.append((key, values))
+        if physical_tag:
+            cell_data["gmsh:physical"].append(
+                physical_tag[0] * numpy.ones(len(values), int)
+            )
+    cells[:] = _gmsh_to_meshio_order(cells)
 
     return cells, cell_data
 
@@ -278,7 +257,7 @@ def _read_periodic(f, is_ascii):
     return periodic
 
 
-def write(filename, mesh, binary=True):
+def write(filename, mesh, float_fmt=".15e", binary=True):
     """Writes msh files, cf.
     <http://gmsh.info//doc/texinfo/gmsh.html#MSH-ASCII-file-format>.
     """
@@ -292,41 +271,32 @@ def write(filename, mesh, binary=True):
         )
 
     if binary:
-        for key, value in mesh.cells.items():
+        for k, (key, value) in enumerate(mesh.cells):
             if value.dtype != c_int:
                 logging.warning(
                     "Binary Gmsh needs c_int (typically numpy.int32) integers "
                     "(got %s). Converting.",
                     value.dtype,
                 )
-                mesh.cells[key] = numpy.array(value, dtype=c_int)
+                mesh.cells[k] = CellBlock(key, numpy.array(value, dtype=c_int))
 
-    # Gmsh cells are mostly ordered like VTK, with a few exceptions:
-    cells = mesh.cells.copy()
-    if "tetra10" in cells:
-        cells["tetra10"] = cells["tetra10"][:, [0, 1, 2, 3, 4, 5, 6, 7, 9, 8]]
-    if "hexahedron20" in cells:
-        cells["hexahedron20"] = cells["hexahedron20"][
-            :, [0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 13, 9, 16, 18, 19, 17, 10, 12, 14, 15]
-        ]
+    cells = _meshio_to_gmsh_order(mesh.cells)
 
     with open(filename, "wb") as fh:
         mode_idx = 1 if binary else 0
         size_of_double = 8
         fh.write(
-            ("$MeshFormat\n4.0 {} {}\n".format(mode_idx, size_of_double)).encode(
-                "utf-8"
-            )
+            "$MeshFormat\n4.0 {} {}\n".format(mode_idx, size_of_double).encode("utf-8")
         )
         if binary:
-            fh.write(struct.pack("i", 1))
-            fh.write("\n".encode("utf-8"))
-        fh.write("$EndMeshFormat\n".encode("utf-8"))
+            numpy.array([1], dtype=c_int).tofile(fh)
+            fh.write(b"\n")
+        fh.write(b"$EndMeshFormat\n")
 
         if mesh.field_data:
             _write_physical_names(fh, mesh.field_data)
 
-        _write_nodes(fh, mesh.points, binary)
+        _write_nodes(fh, mesh.points, float_fmt, binary)
         _write_elements(fh, cells, binary)
         if mesh.gmsh_periodic is not None:
             _write_periodic(fh, mesh.gmsh_periodic, binary)
@@ -336,11 +306,9 @@ def write(filename, mesh, binary=True):
         for name, dat in cell_data_raw.items():
             _write_data(fh, "ElementData", name, dat, binary)
 
-    return
 
-
-def _write_nodes(fh, points, binary):
-    fh.write("$Nodes\n".encode("utf-8"))
+def _write_nodes(fh, points, float_fmt, binary):
+    fh.write(b"$Nodes\n")
 
     # TODO not sure what dimEntity is supposed to say
     dim_entity = 0
@@ -359,7 +327,7 @@ def _write_nodes(fh, points, binary):
         tmp["index"] = 1 + numpy.arange(len(points))
         tmp["x"] = points
         tmp.tofile(fh)
-        fh.write("\n".encode("utf-8"))
+        fh.write(b"\n")
     else:
         # write all points as one big block
         # numEntityBlocks(unsigned long) numNodes(unsigned long)
@@ -372,27 +340,25 @@ def _write_nodes(fh, points, binary):
             )
         )
 
+        fmt = "{} " + " ".join(3 * ["{:" + float_fmt + "}"]) + "\n"
         for k, x in enumerate(points):
             # tag(int) x(double) y(double) z(double)
-            fh.write(
-                "{} {!r} {!r} {!r}\n".format(k + 1, x[0], x[1], x[2]).encode("utf-8")
-            )
+            fh.write(fmt.format(k + 1, x[0], x[1], x[2]).encode("utf-8"))
 
-    fh.write("$EndNodes\n".encode("utf-8"))
-    return
+    fh.write(b"$EndNodes\n")
 
 
 def _write_elements(fh, cells, binary):
     # TODO respect binary
     # write elements
-    fh.write("$Elements\n".encode("utf-8"))
+    fh.write(b"$Elements\n")
 
     if binary:
-        total_num_cells = sum([data.shape[0] for _, data in cells.items()])
+        total_num_cells = sum([data.shape[0] for _, data in cells])
         numpy.array([len(cells), total_num_cells], dtype=c_ulong).tofile(fh)
 
         consecutive_index = 0
-        for cell_type, node_idcs in cells.items():
+        for cell_type, node_idcs in cells:
             # tagEntity(int) dimEntity(int) typeEle(int) numElements(unsigned long)
             numpy.array(
                 [1, _geometric_dimension[cell_type], _meshio_to_gmsh_type[cell_type]],
@@ -416,14 +382,14 @@ def _write_elements(fh, cells, binary):
             data.tofile(fh)
             consecutive_index += len(node_idcs)
 
-        fh.write("\n".encode("utf-8"))
+        fh.write(b"\n")
     else:
         # count all cells
-        total_num_cells = sum([data.shape[0] for _, data in cells.items()])
+        total_num_cells = sum([data.shape[0] for _, data in cells])
         fh.write("{} {}\n".format(len(cells), total_num_cells).encode("utf-8"))
 
         consecutive_index = 0
-        for cell_type, node_idcs in cells.items():
+        for cell_type, node_idcs in cells:
             # tagEntity(int) dimEntity(int) typeEle(int) numElements(unsigned long)
             fh.write(
                 "{} {} {} {}\n".format(
@@ -441,8 +407,7 @@ def _write_elements(fh, cells, binary):
                 fh.write(fmt.format(consecutive_index, *idx).encode("utf-8"))
                 consecutive_index += 1
 
-    fh.write("$EndElements\n".encode("utf-8"))
-    return
+    fh.write(b"$EndElements\n")
 
 
 def _write_periodic(fh, periodic, binary):
@@ -456,7 +421,7 @@ def _write_periodic(fh, periodic, binary):
             fmt = kwargs.pop("fmt", fmt)
             numpy.savetxt(fh, ary, fmt, **kwargs)
 
-    fh.write("$Periodic\n".encode("utf-8"))
+    fh.write(b"$Periodic\n")
     tofile(fh, len(periodic), c_int)
     for dim, (stag, mtag), affine, slave_master in periodic:
         tofile(fh, [dim, stag, mtag], c_int)
@@ -469,72 +434,5 @@ def _write_periodic(fh, periodic, binary):
         tofile(fh, len(slave_master), c_int)
         tofile(fh, slave_master, c_int)
     if binary:
-        fh.write("\n".encode("utf-8"))
-    fh.write("$EndPeriodic\n".encode("utf-8"))
-
-
-_geometric_dimension = {
-    "line": 1,
-    "triangle": 2,
-    "quad": 2,
-    "tetra": 3,
-    "hexahedron": 3,
-    "wedge": 3,
-    "pyramid": 3,
-    "line3": 1,
-    "triangle6": 2,
-    "quad9": 2,
-    "tetra10": 3,
-    "hexahedron27": 3,
-    "wedge18": 3,
-    "pyramid14": 3,
-    "vertex": 0,
-    "quad8": 2,
-    "hexahedron20": 3,
-    "triangle10": 2,
-    "triangle15": 2,
-    "triangle21": 2,
-    "line4": 1,
-    "line5": 1,
-    "line6": 1,
-    "tetra20": 3,
-    "tetra35": 3,
-    "tetra56": 3,
-    "quad16": 2,
-    "quad25": 2,
-    "quad36": 2,
-    "triangle28": 2,
-    "triangle36": 2,
-    "triangle45": 2,
-    "triangle55": 2,
-    "triangle66": 2,
-    "quad49": 2,
-    "quad64": 2,
-    "quad81": 2,
-    "quad100": 2,
-    "quad121": 2,
-    "line7": 1,
-    "line8": 1,
-    "line9": 1,
-    "line10": 1,
-    "line11": 1,
-    "tetra84": 3,
-    "tetra120": 3,
-    "tetra165": 3,
-    "tetra220": 3,
-    "tetra286": 3,
-    "wedge40": 3,
-    "wedge75": 3,
-    "hexahedron64": 3,
-    "hexahedron125": 3,
-    "hexahedron216": 3,
-    "hexahedron343": 3,
-    "hexahedron512": 3,
-    "hexahedron729": 3,
-    "hexahedron1000": 3,
-    "wedge126": 3,
-    "wedge196": 3,
-    "wedge288": 3,
-    "wedge405": 3,
-    "wedge550": 3,
-}
+        fh.write(b"\n")
+    fh.write(b"$EndPeriodic\n")
