@@ -3,16 +3,17 @@ I/O for Gmsh's msh format (version 4.1, as used by Gmsh 4.2.2+), cf.
 <http://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format>.
 """
 import logging
-import struct
 from functools import partial
 
 import numpy
 
-from .._common import cell_data_from_raw, raw_from_cell_data
+from .._common import _geometric_dimension, cell_data_from_raw, raw_from_cell_data
 from .._exceptions import ReadError, WriteError
-from .._mesh import Mesh
+from .._mesh import CellBlock, Mesh
 from .common import (
+    _gmsh_to_meshio_order,
     _gmsh_to_meshio_type,
+    _meshio_to_gmsh_order,
     _meshio_to_gmsh_type,
     _read_data,
     _read_physical_names,
@@ -36,12 +37,12 @@ def read_buffer(f, is_ascii, data_size):
 
     # Initialize the optional data fields
     points = []
-    cells = {}
     field_data = {}
     cell_data_raw = {}
     cell_tags = {}
     point_data = {}
     physical_tags = None
+    cell_sets = {}
     periodic = None
     while True:
         line = f.readline().decode("utf-8")
@@ -54,13 +55,14 @@ def read_buffer(f, is_ascii, data_size):
 
         if environ == "PhysicalNames":
             _read_physical_names(f, field_data)
+            cell_sets = {k: [] for k in field_data.keys()}
         elif environ == "Entities":
             physical_tags = _read_entities(f, is_ascii, data_size)
         elif environ == "Nodes":
             points, point_tags = _read_nodes(f, is_ascii, data_size)
         elif environ == "Elements":
             cells, cell_tags = _read_elements(
-                f, point_tags, physical_tags, is_ascii, data_size
+                f, point_tags, physical_tags, is_ascii, data_size, field_data, cell_sets
             )
         elif environ == "Periodic":
             periodic = _read_periodic(f, is_ascii, data_size)
@@ -96,6 +98,7 @@ def read_buffer(f, is_ascii, data_size):
         point_data=point_data,
         cell_data=cell_data,
         field_data=field_data,
+        cell_sets=cell_sets,
         gmsh_periodic=periodic,
     )
 
@@ -173,7 +176,9 @@ def _read_nodes(f, is_ascii, data_size):
     return points, tags
 
 
-def _read_elements(f, point_tags, physical_tags, is_ascii, data_size):
+def _read_elements(
+    f, point_tags, physical_tags, is_ascii, data_size, field_data, cell_sets
+):
     fromfile = partial(numpy.fromfile, sep=" " if is_ascii else "")
     c_size_t = _size_type(data_size)
 
@@ -189,6 +194,17 @@ def _read_elements(f, point_tags, physical_tags, is_ascii, data_size):
         # entityDim(int) entityTag(int) elementType(int) numElements(size_t)
         dim_entity, tag_entity, type_ele = fromfile(f, c_int, 3)
         (num_ele,) = fromfile(f, c_size_t, 1)
+        for physical_name, cell_set in cell_sets.items():
+            if physical_tags:
+                for physical_tag in physical_tags[dim_entity][tag_entity]:
+                    cell_set.append(
+                        numpy.arange(
+                            num_ele
+                            if field_data[physical_name][0] == physical_tag
+                            else 0,
+                            dtype=type(num_ele),
+                        )
+                    )
         tpe = _gmsh_to_meshio_type[type_ele]
         num_nodes_per_ele = num_nodes_per_cell[tpe]
         d = fromfile(f, c_size_t, int(num_ele * (1 + num_nodes_per_ele))).reshape(
@@ -216,35 +232,16 @@ def _read_elements(f, point_tags, physical_tags, is_ascii, data_size):
         (physical_tag, tpe, inv_tags[d[:, 1:] - 1]) for physical_tag, tpe, d in data
     ]
 
-    cells = {}
+    cells = []
     for physical_tag, key, values in data:
-        if key in cells:
-            cells[key] = numpy.concatenate([cells[key], values])
-            if physical_tag:
-                if key not in cell_data:
-                    cell_data[key] = {}
-                cell_data[key]["gmsh:physical"] = numpy.concatenate(
-                    [
-                        cell_data[key]["gmsh:physical"],
-                        physical_tag[0] * numpy.ones(len(values), int),
-                    ]
-                )
-        else:
-            cells[key] = values
-            if physical_tag:
-                if key not in cell_data:
-                    cell_data[key] = {}
-                cell_data[key]["gmsh:physical"] = physical_tag[0] * numpy.ones(
-                    len(values), int
-                )
-
-    # Gmsh cells are mostly ordered like VTK, with a few exceptions:
-    if "tetra10" in cells:
-        cells["tetra10"] = cells["tetra10"][:, [0, 1, 2, 3, 4, 5, 6, 7, 9, 8]]
-    if "hexahedron20" in cells:
-        cells["hexahedron20"] = cells["hexahedron20"][
-            :, [0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 16, 9, 17, 10, 18, 19, 12, 15, 13, 14]
-        ]
+        cells.append((key, values))
+        if physical_tag:
+            if "gmsh:physical" not in cell_data:
+                cell_data["gmsh:physical"] = []
+            cell_data["gmsh:physical"].append(
+                physical_tag[0] * numpy.ones(len(values), int)
+            )
+    cells[:] = _gmsh_to_meshio_order(cells)
 
     return cells, cell_data
 
@@ -295,31 +292,24 @@ def write4_1(filename, mesh, binary=True):
         )
 
     if binary:
-        for key, value in mesh.cells.items():
+        for k, (key, value) in enumerate(mesh.cells):
             if value.dtype != c_int:
                 logging.warning(
                     "Binary Gmsh needs c_size_t (got %s). Converting.", value.dtype
                 )
-                mesh.cells[key] = value.astype(c_size_t)
+                mesh.cells[k] = CellBlock(key, value.astype(c_size_t))
 
-    # Gmsh cells are mostly ordered like VTK, with a few exceptions:
-    cells = mesh.cells.copy()
-    if "tetra10" in cells:
-        cells["tetra10"] = cells["tetra10"][:, [0, 1, 2, 3, 4, 5, 6, 7, 9, 8]]
-    if "hexahedron20" in cells:
-        cells["hexahedron20"] = cells["hexahedron20"][
-            :, [0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 13, 9, 16, 18, 19, 17, 10, 12, 14, 15]
-        ]
+    cells = _meshio_to_gmsh_order(mesh.cells)
 
     with open(filename, "wb") as fh:
         file_type = 1 if binary else 0
         data_size = c_size_t.itemsize
-        fh.write("$MeshFormat\n".encode("utf-8"))
+        fh.write(b"$MeshFormat\n")
         fh.write("4.1 {} {}\n".format(file_type, data_size).encode("utf-8"))
         if binary:
-            fh.write(struct.pack("i", 1))
-            fh.write("\n".encode("utf-8"))
-        fh.write("$EndMeshFormat\n".encode("utf-8"))
+            numpy.array([1], dtype=c_int).tofile(fh)
+            fh.write(b"\n")
+        fh.write(b"$EndMeshFormat\n")
 
         if mesh.field_data:
             _write_physical_names(fh, mesh.field_data)
@@ -331,6 +321,7 @@ def write4_1(filename, mesh, binary=True):
             _write_periodic(fh, mesh.gmsh_periodic, binary)
         for name, dat in mesh.point_data.items():
             _write_data(fh, "NodeData", name, dat, binary)
+
         cell_data_raw = raw_from_cell_data(mesh.cell_data)
         for name, dat in cell_data_raw.items():
             _write_data(fh, "ElementData", name, dat, binary)
@@ -370,14 +361,14 @@ def _write_entities(fh, cells, binary):
 
 
 def _write_nodes(fh, points, cells, binary):
-    fh.write("$Nodes\n".encode("utf-8"))
+    fh.write(b"$Nodes\n")
 
     # The entity_dim and entity_tag in the $Elements section must correspond to an
     # entity_dim and entity_tag array in the $Nodes section.
     # TODO Not sure what to do if there are multiple element types present.
     if len(cells) != 1:
         raise WriteError("Can only deal with one cell type for now")
-    dim_entity = _geometric_dimension[list(cells.keys())[0]]
+    dim_entity = _geometric_dimension[cells[0][0]]
     entity_tag = 0
 
     # write all points as one big block
@@ -412,7 +403,7 @@ def _write_nodes(fh, points, cells, binary):
         numpy.array([n], dtype=c_size_t).tofile(fh)
         numpy.arange(1, 1 + n, dtype=c_size_t).tofile(fh)
         points.tofile(fh)
-        fh.write("\n".encode("utf-8"))
+        fh.write(b"\n")
     else:
         fh.write(
             "{} {} {} {}\n".format(num_blocks, n, min_tag, max_tag).encode("utf-8")
@@ -423,10 +414,10 @@ def _write_nodes(fh, points, cells, binary):
             )
         )
         numpy.arange(1, 1 + n, dtype=c_size_t).tofile(fh, "\n", "%d")
-        fh.write("\n".encode("utf-8"))
+        fh.write(b"\n")
         numpy.savetxt(fh, points, delimiter=" ")
 
-    fh.write("$EndNodes\n".encode("utf-8"))
+    fh.write(b"$EndNodes\n")
     return
 
 
@@ -442,9 +433,9 @@ def _write_elements(fh, cells, binary):
       ...
     $EndElements
     """
-    fh.write("$Elements\n".encode("utf-8"))
+    fh.write(b"$Elements\n")
 
-    total_num_cells = sum(map(len, cells.values()))
+    total_num_cells = sum(len(c) for _, c in cells)
     num_blocks = len(cells)
     min_element_tag = 1
     max_element_tag = total_num_cells
@@ -455,7 +446,7 @@ def _write_elements(fh, cells, binary):
         ).tofile(fh)
 
         tag0 = 1
-        for cell_type, node_idcs in cells.items():
+        for cell_type, node_idcs in cells:
             # entityDim(int) entityTag(int) elementType(int)
             # numElementsBlock(size_t)
             dim = _geometric_dimension[cell_type]
@@ -481,7 +472,7 @@ def _write_elements(fh, cells, binary):
             ).tofile(fh)
             tag0 += n
 
-        fh.write("\n".encode("utf-8"))
+        fh.write(b"\n")
     else:
         fh.write(
             "{} {} {} {}\n".format(
@@ -490,7 +481,7 @@ def _write_elements(fh, cells, binary):
         )
 
         tag0 = 1
-        for cell_type, node_idcs in cells.items():
+        for cell_type, node_idcs in cells:
             # entityDim(int) entityTag(int) elementType(int) numElementsBlock(size_t)
             dim = _geometric_dimension[cell_type]
             entity_tag = 0
@@ -509,7 +500,7 @@ def _write_elements(fh, cells, binary):
             )
             tag0 += n
 
-    fh.write("$EndElements\n".encode("utf-8"))
+    fh.write(b"$EndElements\n")
     return
 
 
@@ -540,7 +531,7 @@ def _write_periodic(fh, periodic, binary):
             fmt = kwargs.pop("fmt", fmt)
             numpy.savetxt(fh, ary, fmt, **kwargs)
 
-    fh.write("$Periodic\n".encode("utf-8"))
+    fh.write(b"$Periodic\n")
     tofile(fh, len(periodic), c_size_t)
     for dim, (stag, mtag), affine, slave_master in periodic:
         tofile(fh, [dim, stag, mtag], c_int)
@@ -555,72 +546,5 @@ def _write_periodic(fh, periodic, binary):
         tofile(fh, len(slave_master), c_size_t)
         tofile(fh, slave_master, c_size_t)
     if binary:
-        fh.write("\n".encode("utf-8"))
-    fh.write("$EndPeriodic\n".encode("utf-8"))
-
-
-_geometric_dimension = {
-    "line": 1,
-    "triangle": 2,
-    "quad": 2,
-    "tetra": 3,
-    "hexahedron": 3,
-    "wedge": 3,
-    "pyramid": 3,
-    "line3": 1,
-    "triangle6": 2,
-    "quad9": 2,
-    "tetra10": 3,
-    "hexahedron27": 3,
-    "wedge18": 3,
-    "pyramid14": 3,
-    "vertex": 0,
-    "quad8": 2,
-    "hexahedron20": 3,
-    "triangle10": 2,
-    "triangle15": 2,
-    "triangle21": 2,
-    "line4": 1,
-    "line5": 1,
-    "line6": 1,
-    "tetra20": 3,
-    "tetra35": 3,
-    "tetra56": 3,
-    "quad16": 2,
-    "quad25": 2,
-    "quad36": 2,
-    "triangle28": 2,
-    "triangle36": 2,
-    "triangle45": 2,
-    "triangle55": 2,
-    "triangle66": 2,
-    "quad49": 2,
-    "quad64": 2,
-    "quad81": 2,
-    "quad100": 2,
-    "quad121": 2,
-    "line7": 1,
-    "line8": 1,
-    "line9": 1,
-    "line10": 1,
-    "line11": 1,
-    "tetra84": 3,
-    "tetra120": 3,
-    "tetra165": 3,
-    "tetra220": 3,
-    "tetra286": 3,
-    "wedge40": 3,
-    "wedge75": 3,
-    "hexahedron64": 3,
-    "hexahedron125": 3,
-    "hexahedron216": 3,
-    "hexahedron343": 3,
-    "hexahedron512": 3,
-    "hexahedron729": 3,
-    "hexahedron1000": 3,
-    "wedge126": 3,
-    "wedge196": 3,
-    "wedge288": 3,
-    "wedge405": 3,
-    "wedge550": 3,
-}
+        fh.write(b"\n")
+    fh.write(b"$EndPeriodic\n")
