@@ -7,40 +7,52 @@ from functools import reduce
 import numpy
 
 from ..__about__ import __version__
-from .._common import meshio_to_vtk_type, raw_from_cell_data, vtk_to_meshio_type
+from .._common import meshio_to_vtk_type, vtk_to_meshio_type
 from .._exceptions import ReadError, WriteError
 from .._files import open_file
 from .._helpers import register
-from .._mesh import Mesh
+from .._mesh import CellBlock, Mesh
 
-vtk_type_to_numnodes = {
-    0: 0,  # empty
-    1: 1,  # vertex
-    3: 2,  # line
-    5: 3,  # triangle
-    9: 4,  # quad
-    10: 4,  # tetra
-    12: 8,  # hexahedron
-    13: 6,  # wedge
-    14: 5,  # pyramid
-    15: 10,  # penta_prism
-    16: 12,  # hexa_prism
-    21: 3,  # line3
-    22: 6,  # triangle6
-    23: 8,  # quad8
-    24: 10,  # tetra10
-    25: 20,  # hexahedron20
-    26: 15,  # wedge15
-    27: 13,  # pyramid13
-    28: 9,  # quad9
-    29: 27,  # hexahedron27
-    30: 6,  # quad6
-    31: 12,  # wedge12
-    32: 18,  # wedge18
-    33: 24,  # hexahedron24
-    34: 7,  # triangle7
-    35: 4,  # line4
-}
+vtk_type_to_numnodes = numpy.array(
+    [
+        0,  # empty
+        1,  # vertex
+        -1,  # poly_vertex
+        2,  # line
+        -1,  # poly_line
+        3,  # triangle
+        -1,  # triangle_strip
+        -1,  # polygon
+        -1,  # pixel
+        4,  # quad
+        4,  # tetra
+        -1,  # voxel
+        8,  # hexahedron
+        6,  # wedge
+        5,  # pyramid
+        10,  # penta_prism
+        12,  # hexa_prism
+        -1,
+        -1,
+        -1,
+        -1,
+        3,  # line3
+        6,  # triangle6
+        8,  # quad8
+        10,  # tetra10
+        20,  # hexahedron20
+        15,  # wedge15
+        13,  # pyramid13
+        9,  # quad9
+        27,  # hexahedron27
+        6,  # quad6
+        12,  # wedge12
+        18,  # wedge18
+        24,  # hexahedron24
+        7,  # triangle7
+        4,  # line4
+    ]
+)
 
 
 # These are all VTK data types. One sometimes finds 'vtktypeint64', but
@@ -114,11 +126,11 @@ class Info:
         self.is_ascii = False
         self.split = []
         self.num_items = 0
-        # One of the problem in reading VTK files are POINT_DATA and CELL_DATA fields. They
-        # can contain a number of SCALARS+LOOKUP_TABLE tables, without giving and indication
-        # of how many there are. Hence, SCALARS must be treated like a first-class section.
-        # To associate it with POINT/CELL_DATA, we store the `active` section in this
-        # variable.
+        # One of the problem in reading VTK files are POINT_DATA and CELL_DATA fields.
+        # They can contain a number of SCALARS+LOOKUP_TABLE tables, without giving and
+        # indication of how many there are. Hence, SCALARS must be treated like a
+        # first-class section.  To associate it with POINT/CELL_DATA, we store the
+        # `active` section in this variable.
         self.section = None
 
 
@@ -462,6 +474,7 @@ def _read_fields(f, num_fields, is_ascii):
             name, shape0, shape1, data_type = f.readline().decode("utf-8").split()
         else:
             name, shape0, shape1, data_type = line
+
         shape0 = int(shape0)
         shape1 = int(shape1)
         dtype = numpy.dtype(vtk_to_numpy_dtype_name[data_type.lower()])
@@ -497,20 +510,15 @@ def _skip_meta(f):
 
 def translate_cells(data, types, cell_data_raw):
     # https://www.vtk.org/doc/nightly/html/vtkCellType_8h_source.html
-    # Translate it into the cells dictionary.
+    # Translate it into the cells array.
     # `data` is a one-dimensional vector with
     # (num_points0, p0, p1, ... ,pk, numpoints1, p10, p11, ..., p1k, ...
+    has_polygon = numpy.any(types == meshio_to_vtk_type["polygon"])
 
-    # Collect types into bins.
-    # See <https://stackoverflow.com/q/47310359/353337> for better alternatives.
-    bins = {u: numpy.where(types == u)[0] for u in numpy.unique(types)}
-    has_polygon = meshio_to_vtk_type["polygon"] in bins
-
-    # Deduct offsets from the cell types. This is much faster than manually going
-    # through the data array. Slight disadvantage: This doesn't work for cells with a
-    # custom number of points.
-    numnodes = numpy.empty(len(types), dtype=int)
+    cells = []
+    cell_data = {}
     if has_polygon:
+        numnodes = numpy.empty(len(types), dtype=int)
         # If some polygons are in the VTK file, loop over the cells
         nbcells = len(types)
         offsets = numpy.empty(len(types), dtype=int)
@@ -518,19 +526,12 @@ def translate_cells(data, types, cell_data_raw):
         for idx in range(nbcells - 1):
             numnodes[idx] = data[offsets[idx]]
             offsets[idx + 1] = offsets[idx] + numnodes[idx] + 1
+
         idx = nbcells - 1
         numnodes[idx] = data[offsets[idx]]
-    else:
-        for tpe, idx in bins.items():
-            numnodes[idx] = vtk_type_to_numnodes[tpe]
-        offsets = numpy.cumsum(numnodes + 1) - (numnodes + 1)
+        if not numpy.all(numnodes == data[offsets]):
+            raise ReadError()
 
-    if not numpy.all(numnodes == data[offsets]):
-        raise ReadError()
-
-    cells = {}
-    cell_data = {}
-    if has_polygon:
         # TODO: cell_data
         for idx in range(nbcells):
             nbedges = data[offsets[idx]]
@@ -538,26 +539,44 @@ def translate_cells(data, types, cell_data_raw):
             end = start + numnodes[idx]
             cell = data[start:end]
             if nbedges == vtk_type_to_numnodes[meshio_to_vtk_type["triangle"]]:
-                key = "triangle"
+                cell_type = "triangle"
             elif nbedges == vtk_type_to_numnodes[meshio_to_vtk_type["quad"]]:
-                key = "quad"
+                cell_type = "quad"
             else:
-                key = "polygon" + str(nbedges)
-            if key in cells:
-                cells[key] = numpy.vstack([cells[key], cell])
+                cell_type = "polygon" + str(nbedges)
+
+            if len(cells) > 0 and cells[-1].type == cell_type:
+                cells[-1].data.append(cell)
             else:
-                cells[key] = numpy.reshape(cell, (1, -1))
+                cells.append(CellBlock(cell_type, [cell]))
+
+        # convert data to numpy arrays
+        for k, c in enumerate(cells):
+            cells[k] = CellBlock(c.type, numpy.array(c.data))
     else:
-        for tpe, b in bins.items():
-            meshio_type = vtk_to_meshio_type[tpe]
-            n = data[offsets[b[0]]]
-            if not (data[offsets[b]] == n).all():
-                raise ReadError()
-            indices = numpy.add.outer(offsets[b], numpy.arange(1, n + 1))
-            cells[meshio_type] = data[indices]
-            cell_data[meshio_type] = {
-                key: value[b] for key, value in cell_data_raw.items()
-            }
+        # Deduct offsets from the cell types. This is much faster than manually going
+        # through the data array. Slight disadvantage: This doesn't work for cells with
+        # a custom number of points.
+        numnodes = vtk_type_to_numnodes[types]
+        if not numpy.all(numnodes > 0):
+            raise ReadError("File contains cells that meshio cannot handle.")
+        offsets = numpy.cumsum(numnodes + 1) - (numnodes + 1)
+
+        if not numpy.all(numnodes == data[offsets]):
+            raise ReadError()
+
+        b = numpy.concatenate(
+            [[0], numpy.where(types[:-1] != types[1:])[0] + 1, [len(types)]]
+        )
+        for start, end in zip(b[:-1], b[1:]):
+            meshio_type = vtk_to_meshio_type[types[start]]
+            n = data[offsets[start]]
+            indices = numpy.add.outer(offsets[start:end], numpy.arange(1, n + 1))
+            cells.append(CellBlock(meshio_type, data[indices]))
+            for name, d in cell_data_raw.items():
+                if name not in cell_data:
+                    cell_data[name] = []
+                cell_data[name].append(d[start:end])
 
     return cells, cell_data
 
@@ -584,24 +603,23 @@ def write(filename, mesh, binary=True):
                 )
                 mesh.point_data[name] = pad(values)
 
-    if mesh.cell_data:
-        for t, data in mesh.cell_data.items():
-            for name, values in data.items():
-                if len(values.shape) == 2 and values.shape[1] == 2:
-                    logging.warning(
-                        "VTK requires 3D vectors, but 2D vectors given. "
-                        "Appending 0 third component to {}.".format(name)
-                    )
-                    mesh.cell_data[t][name] = pad(mesh.cell_data[t][name])
+    for name, data in mesh.cell_data.items():
+        for k, values in enumerate(data):
+            if len(values.shape) == 2 and values.shape[1] == 2:
+                logging.warning(
+                    "VTK requires 3D vectors, but 2D vectors given. "
+                    "Appending 0 third component to {}.".format(name)
+                )
+                data[k] = pad(data[k])
 
     if not binary:
         logging.warning("VTK ASCII files are only meant for debugging.")
 
     with open_file(filename, "wb") as f:
-        f.write("# vtk DataFile Version 4.2\n".encode("utf-8"))
+        f.write(b"# vtk DataFile Version 4.2\n")
         f.write("written by meshio v{}\n".format(__version__).encode("utf-8"))
         f.write(("BINARY\n" if binary else "ASCII\n").encode("utf-8"))
-        f.write("DATASET UNSTRUCTURED_GRID\n".encode("utf-8"))
+        f.write(b"DATASET UNSTRUCTURED_GRID\n")
 
         # write points and cells
         _write_points(f, points, binary)
@@ -615,10 +633,9 @@ def write(filename, mesh, binary=True):
 
         # write cell data
         if mesh.cell_data:
-            total_num_cells = sum([len(c) for c in mesh.cells.values()])
-            cell_data_raw = raw_from_cell_data(mesh.cell_data)
+            total_num_cells = sum(len(c.data) for c in mesh.cells)
             f.write("CELL_DATA {}\n".format(total_num_cells).encode("utf-8"))
-            _write_field_data(f, cell_data_raw, binary)
+            _write_field_data(f, mesh.cell_data, binary)
 
 
 def _write_points(f, points, binary):
@@ -631,66 +648,70 @@ def _write_points(f, points, binary):
     if binary:
         # Binary data must be big endian, see
         # <https://www.vtk.org/Wiki/VTK/Writing_VTK_files_using_python#.22legacy.22>.
+        # if points.dtype.byteorder == "<" or (
+        #     points.dtype.byteorder == "=" and sys.byteorder == "little"
+        # ):
+        #     logging.warn("Converting to new byte order")
         points.astype(points.dtype.newbyteorder(">")).tofile(f, sep="")
     else:
         # ascii
         points.tofile(f, sep=" ")
-    f.write("\n".encode("utf-8"))
+    f.write(b"\n")
 
 
 def _write_cells(f, cells, binary):
-    total_num_cells = sum([len(c) for c in cells.values()])
-    total_num_idx = sum([numpy.prod(c.shape) for c in cells.values()])
+    total_num_cells = sum([len(c.data) for c in cells])
+    total_num_idx = sum([numpy.prod(c.data.shape) for c in cells])
     # For each cell, the number of nodes is stored
     total_num_idx += total_num_cells
     f.write("CELLS {} {}\n".format(total_num_cells, total_num_idx).encode("utf-8"))
     if binary:
-        for c in cells.values():
-            n = c.shape[1]
+        for c in cells:
+            n = c.data.shape[1]
             dtype = numpy.dtype(">i4")
             # One must force endianness here:
             # <https://github.com/numpy/numpy/issues/15088>
             numpy.column_stack(
-                [numpy.full(c.shape[0], n, dtype=dtype), c.astype(dtype)],
+                [numpy.full(c.data.shape[0], n, dtype=dtype), c.data.astype(dtype)],
             ).astype(dtype).tofile(f, sep="")
-        f.write("\n".encode("utf-8"))
+        f.write(b"\n")
     else:
         # ascii
-        for c in cells.values():
-            n = c.shape[1]
+        for c in cells:
+            n = c.data.shape[1]
             # prepend a column with the value n
-            numpy.column_stack([numpy.full(c.shape[0], n, dtype=c.dtype), c]).tofile(
-                f, sep="\n"
-            )
-            f.write("\n".encode("utf-8"))
+            numpy.column_stack(
+                [numpy.full(c.data.shape[0], n, dtype=c.data.dtype), c.data]
+            ).tofile(f, sep="\n")
+            f.write(b"\n")
 
     # write cell types
     f.write("CELL_TYPES {}\n".format(total_num_cells).encode("utf-8"))
     if binary:
-        for key in cells:
-            key_ = key[:7] if key[:7] == "polygon" else key
+        for c in cells:
+            key_ = c.type[:7] if c.type[:7] == "polygon" else c.type
             vtk_type = meshio_to_vtk_type[key_]
-            numpy.full(len(cells[key]), vtk_type, dtype=numpy.dtype(">i4")).tofile(
+            numpy.full(len(c.data), vtk_type, dtype=numpy.dtype(">i4")).tofile(
                 f, sep=""
             )
-        f.write("\n".encode("utf-8"))
+        f.write(b"\n")
     else:
         # ascii
-        for key in cells:
-            key_ = key[:7] if key[:7] == "polygon" else key
-            numpy.full(len(cells[key]), meshio_to_vtk_type[key_]).tofile(f, sep="\n")
-            f.write("\n".encode("utf-8"))
+        for c in cells:
+            key_ = c.type[:7] if c.type[:7] == "polygon" else c.type
+            numpy.full(len(c.data), meshio_to_vtk_type[key_]).tofile(f, sep="\n")
+            f.write(b"\n")
 
 
 def _write_field_data(f, data, binary):
     f.write(("FIELD FieldData {}\n".format(len(data))).encode("utf-8"))
     for name, values in data.items():
+        if isinstance(values, list):
+            values = numpy.concatenate(values)
         if len(values.shape) == 1:
             num_tuples = values.shape[0]
             num_components = 1
         else:
-            if len(values.shape) != 2:
-                raise WriteError("Only one- and two-dimensional field data supported.")
             num_tuples = values.shape[0]
             num_components = values.shape[1]
 
@@ -715,16 +736,7 @@ def _write_field_data(f, data, binary):
             # ascii
             values.tofile(f, sep=" ")
             # numpy.savetxt(f, points)
-        f.write("\n".encode("utf-8"))
+        f.write(b"\n")
 
 
-register(
-    "vtk",
-    [".vtk"],
-    read,
-    {
-        "vtk": lambda f, m, **kwargs: write(f, m, **kwargs, binary=True),
-        "vtk-ascii": lambda f, m, **kwargs: write(f, m, **kwargs, binary=False),
-        "vtk-binary": lambda f, m, **kwargs: write(f, m, **kwargs, binary=True),
-    },
-)
+register("vtk", [".vtk"], read, {"vtk": write})
