@@ -1,8 +1,11 @@
 """
 I/O for VTU.
+<https://vtk.org/Wiki/VTK_XML_Formats>
+<https://vtk.org/wp-content/uploads/2015/04/file-formats.pdf>
 """
 import base64
 import logging
+import lzma
 import sys
 import zlib
 
@@ -17,7 +20,7 @@ from .._common import (
 )
 from .._exceptions import ReadError
 from .._helpers import register
-from .._mesh import Mesh
+from .._mesh import CellBlock, Mesh
 
 
 def num_bytes_to_num_base64_chars(num_bytes):
@@ -27,22 +30,32 @@ def num_bytes_to_num_base64_chars(num_bytes):
 
 
 def _cells_from_data(connectivity, offsets, types, cell_data_raw):
-    # Translate it into the cells dictionary.
+    # Translate it into the cells array.
     # `connectivity` is a one-dimensional vector with
     # (p0, p1, ... ,pk, p10, p11, ..., p1k, ...
     if len(offsets) != len(types):
         raise ReadError()
 
-    cells = {}
+    b = numpy.concatenate(
+        [[0], numpy.where(types[:-1] != types[1:])[0] + 1, [len(types)]]
+    )
+
+    cells = []
     cell_data = {}
-    for vtk_type in numpy.unique(types):
-        meshio_type = vtk_to_meshio_type[vtk_type]
+    for start, end in zip(b[:-1], b[1:]):
+        try:
+            meshio_type = vtk_to_meshio_type[types[start]]
+        except KeyError:
+            raise ReadError("File contains cells that meshio cannot handle.")
         n = num_nodes_per_cell[meshio_type]
-        # The offsets point to the _end_ of the indices
-        b = types == vtk_type
-        indices = numpy.add.outer(offsets[b], numpy.arange(-n, 0, dtype=offsets.dtype))
-        cells[meshio_type] = connectivity[indices]
-        cell_data[meshio_type] = {key: value[b] for key, value in cell_data_raw.items()}
+        indices = numpy.add.outer(
+            offsets[start:end], numpy.arange(-n, 0, dtype=offsets.dtype)
+        )
+        cells.append(CellBlock(meshio_type, connectivity[indices]))
+        for name, d in cell_data_raw.items():
+            if name not in cell_data:
+                cell_data[name] = []
+            cell_data[name].append(d[start:end])
 
     return cells, cell_data
 
@@ -51,31 +64,15 @@ def _organize_cells(point_offsets, cells, cell_data_raw):
     if len(point_offsets) != len(cells):
         raise ReadError()
 
-    out_cells = {}
-    out_cell_data = {}
+    out_cells = []
     for offset, cls, cdr in zip(point_offsets, cells, cell_data_raw):
         cls, cell_data = _cells_from_data(
             cls["connectivity"], cls["offsets"], cls["types"], cdr
         )
-        for key in cls:
-            if key not in out_cells:
-                out_cells[key] = []
-            out_cells[key].append(cls[key] + offset)
+        for c in cls:
+            out_cells.append(CellBlock(c.type, c.data + offset))
 
-            if key not in out_cell_data:
-                out_cell_data[key] = {}
-
-            for name in cell_data[key]:
-                if name not in out_cell_data[key]:
-                    out_cell_data[key][name] = []
-                out_cell_data[key][name].append(cell_data[key][name])
-
-    for key in out_cells:
-        out_cells[key] = numpy.concatenate(out_cells[key])
-        for name in out_cell_data[key]:
-            out_cell_data[key][name] = numpy.concatenate(out_cell_data[key][name])
-
-    return out_cells, out_cell_data
+    return out_cells, cell_data
 
 
 def get_grid(root):
@@ -126,20 +123,9 @@ class VtuReader:
     """
 
     def __init__(self, filename):  # noqa: C901
-        from lxml import etree as ET
+        import xml.etree.ElementTree as ET
 
-        # libxml2 and with it lxml have a safety net for memory overflows; see, e.g.,
-        # <https://stackoverflow.com/q/33828728/353337>.
-        # This causes the error
-        # ```
-        # cannot parse large files and instead throws the exception
-        #
-        # lxml.etree.XMLSyntaxError: xmlSAX2Characters: huge text node, [...]
-        # ```
-        # Setting huge_tree=True removes the limit. Another alternative would be to use
-        # Python's native xml parser to avoid this error,
-        # import xml.etree.cElementTree as ET
-        parser = ET.XMLParser(remove_comments=True, huge_tree=True)
+        parser = ET.XMLParser()
         tree = ET.parse(filename, parser)
         root = tree.getroot()
 
@@ -152,11 +138,14 @@ class VtuReader:
                 "Unknown VTU file version '{}'.".format(root.attrib["version"])
             )
 
-        if (
-            "compressor" in root.attrib
-            and root.attrib["compressor"] != "vtkZLibDataCompressor"
-        ):
-            raise ReadError()
+        if "compressor" in root.attrib:
+            assert root.attrib["compressor"] in [
+                "vtkLZMADataCompressor",
+                "vtkZLibDataCompressor",
+            ]
+            self.compression = root.attrib["compressor"]
+        else:
+            self.compression = None
 
         self.header_type = (
             root.attrib["header_type"] if "header_type" in root.attrib else "UInt32"
@@ -275,8 +264,27 @@ class VtuReader:
         )
         self.field_data = field_data
 
-    def read_binary(self, data, data_type):
-        # first read the the block size; it determines the size of the header
+    def read_uncompressed_binary(self, data, data_type):
+        byte_string = base64.b64decode(data)
+
+        dtype = vtu_to_numpy_type[self.header_type]
+        if self.byte_order is not None:
+            dtype = dtype.newbyteorder(
+                "<" if self.byte_order == "LittleEndian" else ">"
+            )
+        num_bytes_per_item = numpy.dtype(dtype).itemsize
+        # total_num_bytes = numpy.frombuffer(byte_string[:num_bytes_per_item], dtype)[0]
+
+        # Read the block data; multiple blocks possible here?
+        dtype = vtu_to_numpy_type[data_type]
+        if self.byte_order is not None:
+            dtype = dtype.newbyteorder(
+                "<" if self.byte_order == "LittleEndian" else ">"
+            )
+        return numpy.frombuffer(byte_string[num_bytes_per_item:], dtype=dtype)
+
+    def read_compressed_binary(self, data, data_type):
+        # first read the block size; it determines the size of the header
         dtype = vtu_to_numpy_type[self.header_type]
         if self.byte_order is not None:
             dtype = dtype.newbyteorder(
@@ -288,7 +296,7 @@ class VtuReader:
         num_blocks = numpy.frombuffer(byte_string, dtype)[0]
 
         # read the entire header
-        num_header_items = 3 + num_blocks
+        num_header_items = 3 + int(num_blocks)
         num_header_bytes = num_bytes_per_item * num_header_items
         num_header_chars = num_bytes_to_num_base64_chars(num_header_bytes)
         byte_string = base64.b64decode(data[:num_header_chars])
@@ -312,11 +320,15 @@ class VtuReader:
         byte_offsets[0] = 0
         numpy.cumsum(block_sizes, out=byte_offsets[1:])
 
+        c = {"vtkLZMADataCompressor": lzma, "vtkZLibDataCompressor": zlib}[
+            self.compression
+        ]
+
         # process the compressed data
         block_data = numpy.concatenate(
             [
                 numpy.frombuffer(
-                    zlib.decompress(byte_array[byte_offsets[k] : byte_offsets[k + 1]]),
+                    c.decompress(byte_array[byte_offsets[k] : byte_offsets[k + 1]]),
                     dtype=dtype,
                 )
                 for k in range(num_blocks)
@@ -334,10 +346,20 @@ class VtuReader:
                 c.text, dtype=vtu_to_numpy_type[c.attrib["type"]], sep=" "
             )
         elif fmt == "binary":
-            data = self.read_binary(c.text.strip(), c.attrib["type"])
+            reader = (
+                self.read_uncompressed_binary
+                if self.compression is None
+                else self.read_compressed_binary
+            )
+            data = reader(c.text.strip(), c.attrib["type"])
         elif fmt == "appended":
             offset = int(c.attrib["offset"])
-            data = self.read_binary(self.appended_data[offset:], c.attrib["type"])
+            reader = (
+                self.read_uncompressed_binary
+                if self.compression is None
+                else self.read_compressed_binary
+            )
+            data = reader(self.appended_data[offset:], c.attrib["type"])
         else:
             raise ReadError("Unknown data format '{}'.".format(fmt))
 
@@ -364,11 +386,10 @@ def _chunk_it(array, n):
         k += 1
 
 
-def write(filename, mesh, binary=True):
+def write(filename, mesh, binary=True, compression="zlib", header_type=None):
     # Writing XML with an etree required first transforming the (potentially large)
     # arrays into string, which are much larger in memory still. This makes this writer
     # very memory hungry. See <https://stackoverflow.com/q/59272477/353337>.
-    # from lxml import etree as ET
     from .._cxml import etree as ET
 
     if not binary:
@@ -383,8 +404,6 @@ def write(filename, mesh, binary=True):
             [mesh.points[:, 0], mesh.points[:, 1], numpy.zeros(mesh.points.shape[0])]
         )
 
-    header_type = "UInt32"
-
     vtk_file = ET.Element(
         "VTKFile",
         type="UnstructuredGrid",
@@ -392,9 +411,19 @@ def write(filename, mesh, binary=True):
         # Use the native endianness. Not strictly necessary, but this simplifies things
         # a bit.
         byte_order=("LittleEndian" if sys.byteorder == "little" else "BigEndian"),
-        header_type=header_type,
-        compressor="vtkZLibDataCompressor",
     )
+    header_type = (
+        "UInt32" if header_type is None else vtk_file.set("header_type", header_type)
+    )
+
+    if binary and compression:
+        # TODO lz4, lzma <https://vtk.org/doc/nightly/html/classvtkDataCompressor.html>
+        compressions = {
+            "lzma": "vtkLZMADataCompressor",
+            "zlib": "vtkZLibDataCompressor",
+        }
+        assert compression in compressions
+        vtk_file.set("compressor", compressions[compression])
 
     # swap the data to match the system byteorder
     # Don't use byteswap to make sure that the dtype is changed; see
@@ -403,8 +432,8 @@ def write(filename, mesh, binary=True):
     for key, data in mesh.point_data.items():
         mesh.point_data[key] = data.astype(data.dtype.newbyteorder("="))
     for data in mesh.cell_data.values():
-        for key, dat in data.items():
-            data[key] = dat.astype(dat.dtype.newbyteorder("="))
+        for k, dat in enumerate(data):
+            data[k] = dat.astype(dat.dtype.newbyteorder("="))
     for key, data in mesh.field_data.items():
         mesh.field_data[key] = data.astype(data.dtype.newbyteorder("="))
 
@@ -416,33 +445,47 @@ def write(filename, mesh, binary=True):
             da.set("NumberOfComponents", "{}".format(data.shape[1]))
         if binary:
             da.set("format", "binary")
+            if compression:
+                # compressed write
+                def text_writer(f):
+                    max_block_size = 32768
+                    data_bytes = data.tostring()
 
-            def text_writer(f):
-                max_block_size = 32768
-                data_bytes = data.tostring()
+                    # round up
+                    num_blocks = -int(-len(data_bytes) // max_block_size)
+                    last_block_size = (
+                        len(data_bytes) - (num_blocks - 1) * max_block_size
+                    )
 
-                # round up
-                num_blocks = -int(-len(data_bytes) // max_block_size)
-                last_block_size = len(data_bytes) - (num_blocks - 1) * max_block_size
+                    # It's too bad that we have to keep all blocks in memory. This is
+                    # necessary because the header, written first, needs to know the
+                    # lengths of all blocks. Also, the blocks are encoded _after_ having
+                    # been concatenated.
+                    c = {"lzma": lzma, "zlib": zlib}[compression]
+                    compressed_blocks = [
+                        # This compress is the slowest part of the writer
+                        c.compress(block)
+                        for block in _chunk_it(data_bytes, max_block_size)
+                    ]
 
-                # It's too bad that we have to keep all blocks in memory. This is
-                # necessary because the header, written first, needs to know the lengths
-                # of all blocks. Also, the blocks are encoded _after_ having been
-                # concatenated.
-                compressed_blocks = [
-                    # This zlib.compress is the slowest part of the writer
-                    zlib.compress(block)
-                    for block in _chunk_it(data_bytes, max_block_size)
-                ]
+                    # collect header
+                    header = numpy.array(
+                        [num_blocks, max_block_size, last_block_size]
+                        + [len(b) for b in compressed_blocks],
+                        dtype=vtu_to_numpy_type[header_type],
+                    )
+                    f.write(base64.b64encode(header.tostring()).decode())
+                    f.write(base64.b64encode(b"".join(compressed_blocks)).decode())
 
-                # collect header
-                header = numpy.array(
-                    [num_blocks, max_block_size, last_block_size]
-                    + [len(b) for b in compressed_blocks],
-                    dtype=vtu_to_numpy_type[header_type],
-                )
-                f.write(base64.b64encode(header.tostring()).decode())
-                f.write(base64.b64encode(b"".join(compressed_blocks)).decode())
+            else:
+                # uncompressed write
+                def text_writer(f):
+                    data_bytes = data.tostring()
+                    # collect header
+                    header = numpy.array(
+                        len(data_bytes), dtype=vtu_to_numpy_type[header_type]
+                    )
+                    f.write(base64.b64encode(header.tostring() + data_bytes).decode())
 
         else:
             da.set("format", "ascii")
@@ -465,7 +508,7 @@ def write(filename, mesh, binary=True):
 
     grid = ET.SubElement(vtk_file, "UnstructuredGrid")
 
-    total_num_cells = sum([len(c) for c in mesh.cells.values()])
+    total_num_cells = sum([len(c.data) for c in mesh.cells])
     piece = ET.SubElement(
         grid,
         "Piece",
@@ -482,18 +525,21 @@ def write(filename, mesh, binary=True):
         cls = ET.SubElement(piece, "Cells")
 
         # create connectivity, offset, type arrays
-        connectivity = numpy.concatenate([v.reshape(-1) for v in mesh.cells.values()])
+        connectivity = numpy.concatenate([v.data.reshape(-1) for v in mesh.cells])
 
         # offset (points to the first element of the next cell)
         offsets = [
-            v.shape[1] * numpy.arange(1, v.shape[0] + 1) for v in mesh.cells.values()
+            v.data.shape[1]
+            * numpy.arange(1, v.data.shape[0] + 1, dtype=connectivity.dtype)
+            for v in mesh.cells
         ]
         for k in range(1, len(offsets)):
             offsets[k] += offsets[k - 1][-1]
         offsets = numpy.concatenate(offsets)
+
         # types
         types = numpy.concatenate(
-            [numpy.full(len(v), meshio_to_vtk_type[k]) for k, v in mesh.cells.items()]
+            [numpy.full(len(v), meshio_to_vtk_type[k]) for k, v in mesh.cells]
         )
 
         numpy_to_xml_array(cls, "connectivity", "{:d}", connectivity)
@@ -511,18 +557,8 @@ def write(filename, mesh, binary=True):
             numpy_to_xml_array(cd, name, "{:.11e}", data)
 
     # write_xml(filename, vtk_file, pretty_xml)
-
     tree = ET.ElementTree(vtk_file)
     tree.write(filename)
 
 
-register(
-    "vtu",
-    [".vtu"],
-    read,
-    {
-        "vtu": lambda f, m, **kwargs: write(f, m, **kwargs, binary=True),
-        "vtu-ascii": lambda f, m, **kwargs: write(f, m, **kwargs, binary=False),
-        "vtu-binary": lambda f, m, **kwargs: write(f, m, **kwargs, binary=True),
-    },
-)
+register("vtu", [".vtu"], read, {"vtu": write})
