@@ -3,14 +3,15 @@ I/O for XDMF.
 http://www.xdmf.org/index.php/XDMF_Model_and_Format
 """
 import os
+import xml.etree.ElementTree as ET
 from io import BytesIO
 
 import numpy
 
-from .._common import cell_data_from_raw, raw_from_cell_data, write_xml
+from .._common import CDATA, cell_data_from_raw, raw_from_cell_data, write_xml
 from .._exceptions import ReadError, WriteError
 from .._helpers import register
-from .._mesh import Mesh
+from .._mesh import CellBlock, Mesh
 from .common import (
     attribute_type,
     dtype_to_format_string,
@@ -32,16 +33,14 @@ class XdmfReader:
         self.filename = filename
 
     def read(self):
-        from lxml import etree as ET
-
-        parser = ET.XMLParser(remove_comments=True, huge_tree=True)
+        parser = ET.XMLParser()
         tree = ET.parse(self.filename, parser)
         root = tree.getroot()
 
         if root.tag != "Xdmf":
             raise ReadError()
 
-        version = root.attrib["Version"]
+        version = root.get("Version")
 
         if version.split(".")[0] == "2":
             return self.read_xdmf2(root)
@@ -51,21 +50,28 @@ class XdmfReader:
 
         return self.read_xdmf3(root)
 
-    def _read_data_item(self, data_item):
-        import h5py
+    def _read_data_item(self, data_item, root=None):
+        reference = data_item.get("Reference")
+        if reference:
+            xpath = (data_item.text if reference == "XML" else reference).strip()
+            if xpath.startswith("/"):
+                return self._read_data_item(
+                    root.find(".//" + "/".join(xpath.split("/")[2:])), root
+                )
+            raise ValueError("Can't read XPath {}.".format(xpath))
 
-        dims = [int(d) for d in data_item.attrib["Dimensions"].split()]
+        dims = [int(d) for d in data_item.get("Dimensions").split()]
 
         # Actually, `NumberType` is XDMF2 and `DataType` XDMF3, but many files
         # out there use both keys interchangeably.
-        if "DataType" in data_item.attrib:
-            if "NumberType" in data_item.attrib:
+        if data_item.get("DataType"):
+            if data_item.get("NumberType"):
                 raise ReadError()
-            data_type = data_item.attrib["DataType"]
-        elif "NumberType" in data_item.attrib:
-            if "DataType" in data_item.attrib:
+            data_type = data_item.get("DataType")
+        elif data_item.get("NumberType"):
+            if data_item.get("DataType"):
                 raise ReadError()
-            data_type = data_item.attrib["NumberType"]
+            data_type = data_item.get("NumberType")
         else:
             # Default, see
             # <http://www.xdmf.org/index.php/XDMF_Model_and_Format#XML_Element_.28Xdmf_ClassName.29_and_Default_XML_Attributes>
@@ -76,19 +82,18 @@ class XdmfReader:
         except KeyError:
             precision = "4"
 
-        if data_item.attrib["Format"] == "XML":
-            return numpy.array(
-                data_item.text.split(), dtype=xdmf_to_numpy_type[(data_type, precision)]
+        if data_item.get("Format") == "XML":
+            return numpy.fromstring(
+                data_item.text,
+                dtype=xdmf_to_numpy_type[(data_type, precision)],
+                sep=" ",
             ).reshape(dims)
-        elif data_item.attrib["Format"] == "Binary":
+        elif data_item.get("Format") == "Binary":
             return numpy.fromfile(
                 data_item.text.strip(), dtype=xdmf_to_numpy_type[(data_type, precision)]
             ).reshape(dims)
-
-        if data_item.attrib["Format"] != "HDF":
-            raise ReadError(
-                "Unknown XDMF Format '{}'.".format(data_item.attrib["Format"])
-            )
+        elif data_item.get("Format") != "HDF":
+            raise ReadError("Unknown XDMF Format '{}'.".format(data_item.get("Format")))
 
         info = data_item.text.strip()
         filename, h5path = info.split(":")
@@ -97,22 +102,22 @@ class XdmfReader:
         full_hdf5_path = os.path.join(os.path.dirname(self.filename), filename)
 
         f = h5py.File(full_hdf5_path, "r")
-        if h5path[0] != "/":
-            raise ReadError()
 
-        for key in h5path[1:].split("/"):
+        # Some files don't contain the leading slash /.
+        if h5path[0] == "/":
+            h5path = h5path[1:]
+
+        for key in h5path.split("/"):
             f = f[key]
         # `[()]` gives a numpy.ndarray
         return f[()]
 
     def read_information(self, c_data):
-        from lxml import etree as ET
-
         field_data = {}
         root = ET.fromstring(c_data)
         for child in root:
-            str_tag = child.attrib["key"]
-            dim = int(child.attrib["dim"])
+            str_tag = child.get("key")
+            dim = int(child.get("dim"))
             num_tag = int(child.text)
             field_data[str_tag] = numpy.array([num_tag, dim])
         return field_data
@@ -132,7 +137,7 @@ class XdmfReader:
         if grid.tag != "Grid":
             raise ReadError()
 
-        if "GridType" in grid.attrib and grid.attrib["GridType"] != "Uniform":
+        if grid.get("GridType") not in (None, "Uniform"):
             raise ReadError()
 
         points = None
@@ -146,13 +151,13 @@ class XdmfReader:
                 data_items = list(c)
                 if len(data_items) != 1:
                     raise ReadError()
-                topology_type = c.attrib["TopologyType"]
+                topology_type = c.get("TopologyType")
                 if topology_type == "Mixed":
                     cells = translate_mixed_cells(
                         numpy.fromstring(
                             data_items[0].text,
                             int,
-                            int(data_items[0].attrib["Dimensions"]),
+                            int(data_items[0].get("Dimensions")),
                             " ",
                         )
                     )
@@ -161,7 +166,7 @@ class XdmfReader:
                     cells[meshio_type] = self._read_data_item(data_items[0])
 
             elif c.tag == "Geometry":
-                if "GeometryType" in c.attrib and c.attrib["GeometryType"] != "XYZ":
+                if c.get("GeometryType") not in (None, "XYZ"):
                     raise ReadError()
                 data_items = list(c)
                 if len(data_items) != 1:
@@ -184,14 +189,14 @@ class XdmfReader:
 
                 data = self._read_data_item(data_items[0])
 
-                name = c.attrib["Name"]
-                if c.attrib["Center"] == "Node":
+                name = c.get("Name")
+                if c.get("Center") == "Node":
                     point_data[name] = data
-                elif c.attrib["Center"] == "Cell":
+                elif c.get("Center") == "Cell":
                     cell_data_raw[name] = data
                 else:
                     # TODO field data?
-                    if c.attrib["Center"] != "Grid":
+                    if c.get("Center") != "Grid":
                         raise ReadError()
             else:
                 raise ReadError("Unknown section '{}'.".format(c.tag))
@@ -222,7 +227,7 @@ class XdmfReader:
             raise ReadError()
 
         points = None
-        cells = {}
+        cells = []
         point_data = {}
         cell_data_raw = {}
         field_data = {}
@@ -238,26 +243,28 @@ class XdmfReader:
 
                 # The XDMF2 key is `TopologyType`, just `Type` for XDMF3.
                 # Allow both.
-                if "Type" in c.attrib:
-                    if "TopologyType" in c.attrib:
+                if c.get("Type"):
+                    if c.get("TopologyType"):
                         raise ReadError()
-                    cell_type = c.attrib["Type"]
+                    cell_type = c.get("Type")
                 else:
-                    cell_type = c.attrib["TopologyType"]
+                    cell_type = c.get("TopologyType")
 
                 if cell_type == "Mixed":
                     cells = translate_mixed_cells(data)
                 else:
-                    cells[xdmf_to_meshio_type[cell_type]] = data
+                    cells.append(CellBlock(xdmf_to_meshio_type[cell_type], data))
 
             elif c.tag == "Geometry":
-                try:
-                    geometry_type = c.attrib["GeometryType"]
-                except KeyError:
-                    pass
-                else:
-                    if geometry_type not in ["XY", "XYZ"]:
+                if c.get("Type"):
+                    if c.get("GeometryType"):
                         raise ReadError()
+                    geometry_type = c.get("Type")
+                else:
+                    geometry_type = c.get("GeometryType")
+
+                if geometry_type not in ["XY", "XYZ"]:
+                    raise ReadError('Illegal geometry type "{}".'.format(geometry_type))
 
                 data_items = list(c)
                 if len(data_items) != 1:
@@ -283,11 +290,11 @@ class XdmfReader:
 
                 data = self._read_data_item(data_item)
 
-                name = c.attrib["Name"]
-                if c.attrib["Center"] == "Node":
+                name = c.get("Name")
+                if c.get("Center") == "Node":
                     point_data[name] = data
                 else:
-                    if c.attrib["Center"] != "Cell":
+                    if c.get("Center") != "Cell":
                         raise ReadError()
                     cell_data_raw[name] = data
             else:
@@ -306,16 +313,8 @@ class XdmfReader:
 
 class XdmfWriter:
     def __init__(
-        self,
-        filename,
-        mesh,
-        pretty_xml=True,
-        data_format="HDF",
-        compression=None,
-        compression_opts=None,
+        self, filename, mesh, data_format="HDF", compression="gzip", compression_opts=4
     ):
-        from lxml import etree as ET
-
         if data_format not in ["XML", "Binary", "HDF"]:
             raise WriteError(
                 "Unknown XDMF data format "
@@ -326,11 +325,9 @@ class XdmfWriter:
         self.data_format = data_format
         self.data_counter = 0
         self.compression = compression
-        self.compression_opts = compression_opts
+        self.compression_opts = None if compression is None else compression_opts
 
         if data_format == "HDF":
-            import h5py
-
             self.h5_filename = os.path.splitext(self.filename)[0] + ".h5"
             self.h5_file = h5py.File(self.h5_filename, "w")
 
@@ -350,7 +347,7 @@ class XdmfWriter:
 
         ET.register_namespace("xi", "https://www.w3.org/2001/XInclude/")
 
-        write_xml(filename, xdmf_file, pretty_xml)
+        write_xml(filename, xdmf_file)
 
     def numpy_to_xml_string(self, data):
         if self.data_format == "XML":
@@ -369,7 +366,7 @@ class XdmfWriter:
             return bin_filename
 
         if self.data_format != "HDF":
-            raise WriteError()
+            raise WriteError('Unknown data format "{}"'.format(self.data_format))
         name = "data{}".format(self.data_counter)
         self.data_counter += 1
         self.h5_file.create_dataset(
@@ -381,8 +378,6 @@ class XdmfWriter:
         return os.path.basename(self.h5_filename) + ":/" + name
 
     def points(self, grid, points):
-        from lxml import etree as ET
-
         if points.shape[1] == 1:
             geometry_type = "X"
         elif points.shape[1] == 2:
@@ -406,11 +401,9 @@ class XdmfWriter:
         data_item.text = self.numpy_to_xml_string(points)
 
     def cells(self, cells, grid):
-        from lxml import etree as ET
-
         if len(cells) == 1:
-            meshio_type = list(cells.keys())[0]
-            num_cells = len(cells[meshio_type])
+            meshio_type = cells[0].type
+            num_cells = len(cells[0].data)
             xdmf_type = meshio_to_xdmf_type[meshio_type][0]
             topo = ET.SubElement(
                 grid,
@@ -418,8 +411,8 @@ class XdmfWriter:
                 TopologyType=xdmf_type,
                 NumberOfElements=str(num_cells),
             )
-            dt, prec = numpy_to_xdmf_dtype[cells[meshio_type].dtype.name]
-            dim = "{} {}".format(*cells[meshio_type].shape)
+            dt, prec = numpy_to_xdmf_dtype[cells[0].data.dtype.name]
+            dim = "{} {}".format(*cells[0].data.shape)
             data_item = ET.SubElement(
                 topo,
                 "DataItem",
@@ -428,34 +421,34 @@ class XdmfWriter:
                 Format=self.data_format,
                 Precision=prec,
             )
-            data_item.text = self.numpy_to_xml_string(cells[meshio_type])
-        elif len(cells) > 1:
+            data_item.text = self.numpy_to_xml_string(cells[0].data)
 
-            total_num_cells = sum(c.shape[0] for c in cells.values())
+        else:
+            assert len(cells) > 1
+            total_num_cells = sum(c.data.shape[0] for c in cells)
             topo = ET.SubElement(
                 grid,
                 "Topology",
                 TopologyType="Mixed",
                 NumberOfElements=str(total_num_cells),
             )
-            total_num_cell_items = sum(numpy.prod(c.shape) for c in cells.values())
-            dim = str(
-                total_num_cell_items
-                + total_num_cells
-                + (cells["line"].shape[0] if "line" in cells else 0)
+            total_num_cell_items = sum(numpy.prod(c.data.shape) for c in cells)
+            num_vertices_and_lines = sum(
+                c.data.shape[0] for c in cells if c.type in {"vertex", "line"}
             )
+            dim = str(total_num_cell_items + total_num_cells + num_vertices_and_lines)
             cd = numpy.concatenate(
                 [
                     numpy.hstack(
                         [
                             numpy.full(
-                                (value.shape[0], 2 if key == "line" else 1),
+                                (value.shape[0], 2 if key in {"vertex", "line"} else 1),
                                 meshio_type_to_xdmf_index[key],
                             ),
                             value,
                         ]
                     ).flatten()
-                    for key, value in cells.items()
+                    for key, value in cells
                 ]
             )
             dt, prec = numpy_to_xdmf_dtype[cd.dtype.name]
@@ -470,8 +463,6 @@ class XdmfWriter:
             data_item.text = self.numpy_to_xml_string(cd)
 
     def point_data(self, point_data, grid):
-        from lxml import etree as ET
-
         for name, data in point_data.items():
             att = ET.SubElement(
                 grid,
@@ -493,8 +484,6 @@ class XdmfWriter:
             data_item.text = self.numpy_to_xml_string(data)
 
     def cell_data(self, cell_data, grid):
-        from lxml import etree as ET
-
         raw = raw_from_cell_data(cell_data)
         for name, data in raw.items():
             att = ET.SubElement(
@@ -517,31 +506,25 @@ class XdmfWriter:
             data_item.text = self.numpy_to_xml_string(data)
 
     def field_data(self, field_data, information):
-        from lxml import etree as ET
-
         info = ET.Element("main")
         for name, data in field_data.items():
             data_item = ET.SubElement(info, "map", key=name, dim=str(data[1]))
             data_item.text = str(data[0])
-        information.text = ET.CDATA(ET.tostring(info))
+        # information.text = ET.CDATA(ET.tostring(info))
+        information.append(CDATA(ET.tostring(info).decode("utf-8")))
 
 
 def write(*args, **kwargs):
     XdmfWriter(*args, **kwargs)
 
 
-register(
-    "xdmf",
-    [".xdmf", ".xmf"],
-    read,
-    {
-        "xdmf": write,
-        "xdmf-binary": lambda f, m, **kwargs: write(f, m, data_format="Binary"),
-        "xdmf-hdf": lambda f, m, **kwargs: write(f, m, data_format="HDF"),
-        "xdmf-xml": lambda f, m, **kwargs: write(f, m, data_format="XML"),
-        "xdmf3": write,
-        "xdmf3-binary": lambda f, m, **kwargs: write(f, m, data_format="Binary"),
-        "xdmf3-hdf": lambda f, m, **kwargs: write(f, m, data_format="HDF"),
-        "xdmf3-xml": lambda f, m, **kwargs: write(f, m, data_format="XML"),
-    },
-)
+try:
+    import h5py
+# Use ModuleNotFoundError when dropping support for Python 3.5
+except ImportError:
+    pass
+else:
+    # TODO register all xdmf except hdf outside this try block
+    register(
+        "xdmf", [".xdmf", ".xmf"], read, {"xdmf": write},
+    )
