@@ -15,10 +15,9 @@ import meshio
         helpers.hex_mesh,
     ],
 )
-@pytest.mark.parametrize("write_binary", [False, True])
-def test(mesh, write_binary):
+@pytest.mark.parametrize("binary", [False, True])
+def test(mesh, binary):
     def writer(*args, **kwargs):
-        return meshio._ansys.write(*args, write_binary=write_binary, **kwargs)
+        return meshio.ansys.write(*args, binary=binary, **kwargs)
 
-    helpers.write_read(writer, meshio._ansys.read, mesh, 1.0e-15)
-    return
+    helpers.write_read(writer, meshio.ansys.read, mesh, 1.0e-15)
