@@ -4,8 +4,6 @@ import pytest
 import helpers
 import meshio
 
-lxml = pytest.importorskip("lxml")
-
 
 @pytest.mark.parametrize(
     "mesh",
@@ -13,8 +11,9 @@ lxml = pytest.importorskip("lxml")
         helpers.tri_mesh,
         helpers.tri_mesh_2d,
         helpers.tet_mesh,
-        helpers.add_cell_data(helpers.tri_mesh, 1, dtype=float),
-        helpers.add_cell_data(helpers.tri_mesh, 1, dtype=numpy.int32),
+        helpers.add_cell_data(
+            helpers.tri_mesh, [("a", (), float), ("b", (), numpy.int64)]
+        ),
     ],
 )
 def test_dolfin(mesh):
