@@ -15,9 +15,9 @@ import meshio
         # helpers.tri_quad_mesh,
         helpers.add_point_data(helpers.tri_mesh, 1, dtype=int),
         helpers.add_point_data(helpers.tri_mesh, 1, dtype=float),
-        # helpers.add_cell_data(helpers.tri_mesh, 1),
-        # helpers.add_cell_data(helpers.tri_mesh, 3),
-        # helpers.add_cell_data(helpers.tri_mesh, 9),
+        # helpers.add_cell_data(helpers.tri_mesh, [("a", (), numpy.float64)]),
+        # helpers.add_cell_data(helpers.tri_mesh, [("a", (2,), numpy.float64)]),
+        # helpers.add_cell_data(helpers.tri_mesh, [("a", (3,), numpy.float64)]),
     ],
 )
 @pytest.mark.parametrize("binary", [False, True])
@@ -25,8 +25,8 @@ def test_ply(mesh, binary):
     def writer(*args, **kwargs):
         return meshio.ply.write(*args, binary=binary, **kwargs)
 
-    for key in mesh.cells:
-        mesh.cells[key] = mesh.cells[key].astype(numpy.int32)
+    for k, c in enumerate(mesh.cells):
+        mesh.cells[k] = meshio.CellBlock(c.type, c.data.astype(numpy.int32))
 
     helpers.write_read(writer, meshio.ply.read, mesh, 1.0e-12)
 
@@ -42,4 +42,4 @@ def test_reference_file(filename, ref_sum, ref_num_cells):
     tol = 1.0e-2
     s = numpy.sum(mesh.points)
     assert abs(s - ref_sum) < tol * abs(ref_sum)
-    assert len(mesh.cells["triangle"]) == ref_num_cells
+    assert len(mesh.get_cells_type("triangle")) == ref_num_cells
