@@ -6,9 +6,6 @@ import pytest
 import helpers
 import meshio
 
-lxml = pytest.importorskip("lxml")
-
-
 test_set = [helpers.line_mesh, helpers.tri_mesh, helpers.tri_mesh_2d, helpers.quad_mesh]
 
 
@@ -17,4 +14,3 @@ def test(mesh):
     with tempfile.TemporaryDirectory() as temp_dir:
         filepath = os.path.join(temp_dir, "out.svg")
         meshio.write_points_cells(filepath, mesh.points, mesh.cells)
-    return
