@@ -9,6 +9,5 @@ test_set = [helpers.tet_mesh]
 @pytest.mark.parametrize("mesh", test_set)
 def test(mesh):
     helpers.write_read(
-        meshio._tetgen.write, meshio._tetgen.read, mesh, 1.0e-15, extension=".node"
+        meshio.tetgen.write, meshio.tetgen.read, mesh, 1.0e-15, extension=".node"
     )
-    return
