@@ -24,10 +24,13 @@ test_set = [
     helpers.add_point_data(helpers.tri_mesh, 1),
     helpers.add_point_data(helpers.tri_mesh, 2),
     helpers.add_point_data(helpers.tri_mesh, 3),
-    helpers.add_cell_data(helpers.tri_mesh, 1),
-    helpers.add_cell_data(helpers.tri_mesh, 2),
-    helpers.add_cell_data(helpers.tri_mesh, 3),
-    helpers.add_cell_data(helpers.add_point_data(helpers.tri_mesh_2d, 2), 2),
+    # VTK files float data is always stored in big endian
+    helpers.add_cell_data(helpers.tri_mesh, [("a", (), ">f8")]),
+    helpers.add_cell_data(helpers.tri_mesh, [("a", (2,), ">f8")]),
+    helpers.add_cell_data(helpers.tri_mesh, [("a", (3,), ">f8")]),
+    helpers.add_cell_data(
+        helpers.add_point_data(helpers.tri_mesh_2d, 2), [("a", (2,), ">f8")]
+    ),
 ]
 
 
@@ -58,7 +61,8 @@ def test_reference_file(filename, ref_sum, ref_num_cells, binary):
     tol = 1.0e-2
     s = numpy.sum(mesh.points)
     assert abs(s - ref_sum) < tol * ref_sum
-    assert len(mesh.cells["triangle"]) == ref_num_cells
+    assert mesh.cells[0].type == "triangle"
+    assert len(mesh.cells[0].data) == ref_num_cells
     writer = partial(meshio.vtk.write, binary=binary)
     helpers.write_read(writer, meshio.vtk.read, mesh, 1.0e-15)
 
@@ -80,8 +84,8 @@ def test_structured(filename, ref_cells, ref_num_cells, ref_num_pnt):
 
     mesh = meshio.read(filename)
     assert len(mesh.cells) == 1
-    assert ref_cells in mesh.cells
-    assert len(mesh.cells[ref_cells]) == ref_num_cells
+    assert ref_cells == mesh.cells[0].type
+    assert len(mesh.cells[0].data) == ref_num_cells
     assert len(mesh.points) == ref_num_pnt
 
 
